﻿#r "FSharp.Data.TypeProviders.dll"
#r "System.Data.Linq.dll"
#r @"C:\Users\linq\Documents\Visual Studio 2013\Projects\Demos\packages\FSharpComposableQuery.1.0.3-beta\lib\net40\FSharpComposableQuery.dll"

open Microsoft.FSharp.Data.TypeProviders
open Microsoft.FSharp.Quotations
//open FSharpComposableQuery

open System.Linq

type dbSchema = SqlDataConnection< ConnectionString="Data Source=LINQ-BOX\SQLEXPRESS;Initial Catalog=organisation;Integrated Security=SSPI;">
    
let dbquery = FSharpComposableQuery.TopLevelValues.query


// Schema declarations
type Department = dbSchema.ServiceTypes.Departments

// TypeProvider type abbreviations. 
type Employee = dbSchema.ServiceTypes.Employees

type Contact = dbSchema.ServiceTypes.Contacts

type Task = dbSchema.ServiceTypes.Tasks
    
//Nested type declarations
type EmployeeTasks =
    { emp : string
      tasks : System.Linq.IQueryable<string> }

type DepartmentEmployees =
    { dpt : string
      employees : System.Linq.IQueryable<EmployeeTasks> }

type NestedOrg = System.Linq.IQueryable<DepartmentEmployees>

let db = dbSchema.GetDataContext()

let force xs = xs |> Seq.iter (fun _ -> ())

db.DataContext.Log <- stdout

[<Literal>]
let N_DEPARTMENTS = 40

[<Literal>]
let N_EMPLOYEES = 5000

[<Literal>]
let N_ABSTRACTION = 500

//db tables
let departments = db.Departments
let employees = db.Employees
let contacts = db.Contacts
let tasks = db.Tasks



(*
Example 8 and 9 demonstrate the benefit of using intermediate nested structures. 
Each of them evaluates to the same query but the way we formulate them is inherently different. 

"List all departments where every employee can perform a given task t"
*)

// Example 8

let expertiseNaive =
    <@ fun u ->
        query {
            for d in db.Departments do
                if not (query {
                            for e in db.Employees do
                                exists (e.Dpt = d.Dpt && not (query {
                                                                    for t in db.Tasks do
                                                                        exists (e.Emp = t.Emp && t.Tsk = u)
                                                                }))
                        })
                then yield d
        } @>

let ex8 = query { yield! (%expertiseNaive) "abstract" } |> force

let ex8' = dbquery { yield! (%expertiseNaive) "abstract" } |> force

// Example 9

let nestedDb =
    <@ query {
            for d in db.Departments do
                yield { dpt = d.Dpt
                        employees =
                            query {
                                for e in db.Employees do
                                    if d.Dpt = e.Dpt then
                                        yield { emp = e.Emp
                                                tasks =
                                                    query {
                                                        for t in db.Tasks do
                                                            if t.Emp = e.Emp then yield t.Tsk
                                                    } }
                            } }
        } @>

let any() =
    <@ fun xs p ->
        query {
            for x in xs do
                exists (p x)
        } @>
                

(* There are a number of ways to write each of the following queries *)

let forall() = <@ fun xs p -> not ((%any()) xs (fun x -> not(p x))) @>
        

let contains'() = <@ fun xs u -> (%any()) xs (fun x -> x = u) @>


let expertise =
    <@ fun u ->
        query {
            for d in (%nestedDb) do
                if (%forall()) (d.employees) (fun e -> (%contains'()) e.tasks u) then yield d.dpt
        } @>

let ex9 = query { yield! (%expertise) "abstract" } |> force

let ex9' = dbquery { yield! (%expertise) "abstract" } |> force
