﻿#r "FSharp.Data.TypeProviders.dll"
#r "System.Data.Linq.dll"
#r @"C:\Users\linq\Documents\Visual Studio 2013\Projects\Demos\packages\FSharpComposableQuery.1.0.3-beta\lib\net40\FSharpComposableQuery.dll"

open Microsoft.FSharp.Data.TypeProviders
open Microsoft.FSharp.Quotations
//open FSharpComposableQuery
open System.Linq

[<Literal>]
let dbConfigPath = "db.config"
    
type dbSchemaPeople = SqlDataConnection< ConnectionString="Data Source=LINQ-BOX\SQLEXPRESS;Initial Catalog=MyPeople;Integrated Security=SSPI;">

type Couple = dbSchemaPeople.ServiceTypes.Couples

type Person = dbSchemaPeople.ServiceTypes.People


let db = dbSchemaPeople.GetDataContext()

let force xs = xs |> Seq.iter (fun _ -> ())

let dbquery = FSharpComposableQuery.TopLevelValues.query


db.DataContext.Log <- stdout

// Example 0: Age differences
type Result = { Name : string; Diff : int }

let ex0 =
    query {
        for c in db.Couples do
        for w in db.People do
        for m in db.People do
            if c.Her = w.Name && c.Him = m.Name && w.Age > m.Age then
                yield { Name = w.Name
                        Diff = w.Age - m.Age }
        } |> force

let ex0' =
    dbquery {
        for c in db.Couples do
        for w in db.People do
        for m in db.People do
            if c.Her = w.Name && c.Him = m.Name && w.Age > m.Age then
                yield { Name = w.Name
                        Diff = w.Age - m.Age }
        } |> force

// Example 1: Age differences by reusing a simple "view"

type Ages = {Her:string; Him:string; HerAge:int; HisAge:int }

let view = 
  <@ fun p -> query {  
       for c in db.Couples do
       for w in db.People do
       for m in db.People do
         if c.Her = w.Name && c.Him = m.Name && p w.Name m.Name w.Age m.Age then
           yield { Her = w.Name
                   Him = m.Name
                   HerAge = w.Age
                   HisAge = m.Age
                 }
   }  @>

let ex1 = 
     query {
        for x in (%view) (fun wname mname wage mage -> wage > mage) do 
        yield { Name = x.Her
                Diff = x.HerAge - x.HisAge }
        } |> force

let ex1' = 
     dbquery {
        for x in (%view) (fun wname mname wage mage -> wage > mage) do 
        yield { Name = x.Her
                Diff = x.HerAge - x.HisAge }
        } |> force
        
// Example 2: parameterizing by integers
let rangeSimple =
    fun (a : int) (b : int) ->
        query {
            for u in db.People do
                if a <= u.Age && u.Age < b then
                    yield u
        }

let ex2 = query { yield! rangeSimple 30 40 } |> force

let ex2' = dbquery { yield! rangeSimple 30 40 } |> force

// Example 3, 4: parameterizing by predicates
let satisfies =
    <@ fun p ->
        query {
            for u in db.People do
                if p u.Age then
                    yield u
        } @>

let ex3 = query { yield! (%satisfies) (fun x -> 20 <= x && x < 30) } |> force

let ex3' = dbquery { yield! (%satisfies) (fun x -> 20 <= x && x < 30) } |> force

let ex4 = query { yield! (%satisfies) (fun x -> x % 2 = 0) } |> force

let ex4' = dbquery { yield! (%satisfies) (fun x -> x % 2 = 0) } |> force

// Example 5
let range = 
    <@ fun (a : int) (b : int) ->
        query {
            for u in db.People do
                if a <= u.Age && u.Age < b then
                    yield u
        } @>

let ageFromName =
    <@ fun s ->
        query {
            for u in db.People do
                if s = u.Name then yield u.Age
        } @>

let compose : Expr<string -> string -> IQueryable<Person>> =
    <@ fun s t ->
        query {
            for a in (%ageFromName) s do
                for b in (%ageFromName) t do
                    yield! (%range) a b
        } @>

let ex5 = query { yield! (%compose) "Eve" "Bob" } |> force

let ex5' = dbquery { yield! (%compose) "Eve" "Bob" } |> force

// Example 6, 7

type Predicate =
    | Above of int
    | Below of int
    | And of Predicate * Predicate
    | Or of Predicate * Predicate
    | Not of Predicate

let rec eval (t : Predicate) : Expr<int -> bool> =
    match t with
    | Above n -> <@ fun x -> x >= n @>
    | Below n -> <@ fun x -> x < n @>
    | And(t1, t2) -> <@ fun x -> (%eval t1) x && (%eval t2) x @>
    | Or(t1, t2) -> <@ fun x -> (%eval t1) x || (%eval t2) x @>
    | Not(t0) -> <@ fun x -> not ((%eval t0) x) @>

let ex6 = query { yield! (%satisfies) (%eval (And(Above 20, Below 30))) } |> force

let ex6' = dbquery { yield! (%satisfies) (%eval (And(Above 20, Below 30))) } |> force

let ex7 = query { yield! (%satisfies) (%eval (Not(Or(Below 20, Above 30)))) } |> force

let ex7' = dbquery { yield! (%satisfies) (%eval (And(Not(Or(Below 20, Above 30)),And(Above 20, Below 30)))) } |> force

        
